/**
 *  Copyright (C) Bastiaan van den Berg
 *
 *  This file is part of RNAHairpinDesigner.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * Enumeration of the possible base pairs - two nucleotides on opposite 
 * complementary RNA strands that are connected via hydrogen bonds.
 * <p>
 * The six basepairs are:
 * <pre>
 * 5' g c g u a u  3'
 *    | | | | | |
 * 3' c g u g u a  5'
 * </pre>
 * The base pair GC:
 * <pre>
 * 5' g  3'
 *    |  
 * 3' c  5'
 * </pre>
 * So the first nucleotide is the top one and the second the bottom one.
 * <p>
 * See also http://2008.igem.org/Team:TUDelft/Temperature_software
 * 
 * @see Nucleotide
 * @author Bastiaan van den Berg
 */
public enum BasePair {
	
	GC(Nucleotide.G, Nucleotide.C),
	CG(Nucleotide.C, Nucleotide.G),
	GU(Nucleotide.G, Nucleotide.U),
	UG(Nucleotide.U, Nucleotide.G),
	AU(Nucleotide.A, Nucleotide.U),
	UA(Nucleotide.U, Nucleotide.A);
	
	/**
	 * Creates a new base pair with the given top and bottom nucleotide.
	 * @param top The nucleotide at the top.
	 * @param bottom The nucleotide at the bottom.
	 */
	BasePair(Nucleotide top, Nucleotide bottom) {
		this.top = top;
		this.bottom = bottom;
	}
	
	/**
	 * Returns the nucleotide at the top of the base pair.
	 * @return The top nucleotide.
	 */
	public Nucleotide getTop() {
		return top;
	}
	
	/**
	 * Returns the nucleotide at the bottom of the base pair.
	 * @return The bottom nucleotide.
	 */
	public Nucleotide getBottom() {
		return bottom;
	}
	
	private Nucleotide top;
	private Nucleotide bottom;
}
