/**
 *  Copyright (C) Bastiaan van den Berg
 *
 *  This file is part of RNAHairpinDesigner.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.EnumSet;
import java.util.ArrayList;
import java.util.regex.*;

/**
 * Enumeration of RNA structural loop elements. This enumeration only contains
 * stack loops. A loop consists of a left and a right base pair. The first two
 * letters represent the left base pair and the last two letters represent the second
 * base pair (notice that the second base pair is upside down). 
 * <pre>
 * GCAU = 5'-G-U- 3'
 *           | |  
 *        3'-C-A- 5'
 * </pre>
 * 
 * Each loop has a free energy which have been retrieved from the program 
 * RNAfold.
 * <p>
 * See also http://2008.igem.org/Team:TUDelft/Temperature_software
 * 
 * @see BasePair
 * @see Nucleotide
 * @author Bastiaan van den Berg 
 */
public enum Loop {

    CGCG(BasePair.CG, BasePair.CG, -240, 1),
    CGGC(BasePair.CG, BasePair.GC, -330, 1),
    CGGU(BasePair.CG, BasePair.GU, -210, 1),
    CGUG(BasePair.CG, BasePair.UG, -140, 1),
    CGAU(BasePair.CG, BasePair.AU, -210, 1),
    CGUA(BasePair.CG, BasePair.UA, -210, 1),
    
    GCCG(BasePair.GC, BasePair.CG, -330, 1),
    GCGC(BasePair.GC, BasePair.GC, -340, 1),
    GCGU(BasePair.GC, BasePair.GU, -250, 1),
    GCUG(BasePair.GC, BasePair.UG, -150, 1),
    GCAU(BasePair.GC, BasePair.AU, -220, 1),
    GCUA(BasePair.GC, BasePair.UA, -240, 1),
    
    GUCG(BasePair.GU, BasePair.CG, -210, 1),
    GUGC(BasePair.GU, BasePair.GC, -250, 1),
    GUGU(BasePair.GU, BasePair.GU,  130, 1),
    GUUG(BasePair.GU, BasePair.UG,  -50, 1),
    GUAU(BasePair.GU, BasePair.AU, -140, 1),
    GUUA(BasePair.GU, BasePair.UA, -130, 1),
    
    UGCG(BasePair.UG, BasePair.CG, -140, 1),
    UGGC(BasePair.UG, BasePair.GC, -150, 1),
    UGGU(BasePair.UG, BasePair.GU,  -50, 1),
    UGUG(BasePair.UG, BasePair.UG,   30, 1),
    UGAU(BasePair.UG, BasePair.AU,  -60, 1),
    UGUA(BasePair.UG, BasePair.UA, -100, 1),
    
    AUCG(BasePair.AU, BasePair.CG, -210, 1),
    AUGC(BasePair.AU, BasePair.GC, -220, 1),
    AUGU(BasePair.AU, BasePair.GU, -140, 1),
    AUUG(BasePair.AU, BasePair.UG,  -60, 1),
    AUAU(BasePair.AU, BasePair.AU, -110, 1),
    AUUA(BasePair.AU, BasePair.UA,  -90, 1),
    
    UACG(BasePair.UA, BasePair.CG, -210, 1),
    UAGC(BasePair.UA, BasePair.GC, -240, 1),
    UAGU(BasePair.UA, BasePair.GU, -130, 1),
    UAUG(BasePair.UA, BasePair.UG, -100, 1),
    UAAU(BasePair.UA, BasePair.AU,  -90, 1),
    UAUA(BasePair.UA, BasePair.UA, -130, 1);
    
    /**
     * Creates a new loop.
     * @param left Left base pair.
     * @param right Right base pair.
     * @param energy Free energy.
     * @param size Size of the loop.
     */
    private Loop(BasePair left, BasePair right, int energy, int size) {
        this.left = left;
        this.right = right;
        this.energy = energy;
        this.size = size;
    }
    
    /**
     * This method returns all the Loops that 'fit' the given template and 
     * have an energy value in between the given upper and lower bound.
     * Template has length 4 and contains [cgua.] and lowerbound < upperbound.
     * @param template Template which should fit the loop.
     * @param lowerBound Minimal allowed free energy.
     * @param upperBound Maximal allowed free erergy.
     * @return List with loops that meet the requirements.
     */
    public static ArrayList<Loop> getLoops(String template, int lowerBound, int upperBound) {
        ArrayList<Loop> result = new ArrayList<Loop>();
        Pattern pattern = Pattern.compile(template, Pattern.CASE_INSENSITIVE);
        for(Loop loop : EnumSet.allOf(Loop.class)) {
            Matcher matcher = pattern.matcher(loop.name().toLowerCase());
            int e = loop.getEnergy();
            if(matcher.find() && e >= lowerBound && e <= upperBound) {
                result.add(loop);
            }
        }
        return result;
    }
    
    /**
     * Returns the free energy of the loop.
     * @return Free energy of the loop.
     */
    public int getEnergy() {
        return energy;
    }
    
    /**
     * Returns the size of the loop.
     * @return Size of the loop.
     */
    public int getSize() {
    	return size;
    }
    
    /**
     * Returns the left base pair.
     * @return Left base pair.
     */
    public BasePair getLeft() {
    	return left;
    }
    
    /**
     * Returns the right base pair.
     * @return Right base pair.
     */
    public BasePair getRight() {
    	return right;
    }
    
    /**
     * Returns a string representation of the loop.
     */
    public String toString() {
        return energy + " " + size + "\n";
    }
    
    private BasePair left;
    private BasePair right;
    private int energy;
    private int size;
}
