/**
 *  Copyright (C) Bastiaan van den Berg
 *
 *  This file is part of RNAHairpinDesigner.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;

/**
 * This class represents a node that is part of a tree data structure. A node 
 * represents an RNA loop structure that is part of a hairpin structure. A branch
 * within the tree represents a hairpin structure that is build up by a number
 * of loops, which are the nodes.
 * <p>
 * See also http://2008.igem.org/Team:TUDelft/Temperature_software
 * 
 * @see Tree
 * @author Bastiaan van den Berg 
 */
public class Node {

    /**
     * Constructs a new Node.
     * @param tree The tree to which this node belongs.
     * @param parent The parent node of this node.
     * @param level The level on which this node resides within the tree.
     * @param loop The loop that this node represents.
     * @param energy The stacked energy from the root node to this node (the
     *               sum of all the energy values of the loops from this node
     *               to the root node)
     */
    public Node(Tree tree, Node parent, int level, Loop loop, int energy) {
        this.tree = tree;
        this.level = level;
        this.parent = parent;
        this.loop = loop;
        this.stackedEnergy = energy;
    }
    
    /**
     * This method builds up a whole tree by recursively adding childs. All the 
     * loops that fit the template, fit the loop represented by this node (the 
     * loop is added to the right of this loop so the left base pair should be
     * the same as the right base pair of this nodes loop), and that provide a 
     * hairpin with a stability profile that stays within the boundaries, are 
     * added as childs to this node. This method is then applied to all the just
     * added childs.
     */
    public void addChilds() {
        
        // build template
        String template = buildChildTemplate();
        
        // determine upper and lower bound of the energy based on the energy
        // profile and the maximal distance in tree
        int[] profile = tree.getEnergyProfile();
        int lowBound = (profile[level] - tree.getMaxDist()) - stackedEnergy;
        int uppBound = (profile[level] + tree.getMaxDist()) - stackedEnergy;
        
        // get possible child stacks, the ones that fit the template and have 
        // an energy value in between the lower and upper bound
        ArrayList<Loop> loops = Loop.getLoops(template, lowBound, uppBound);
        
        // add childs
        for(Loop loop : loops) {
            int newStackedEnergy = stackedEnergy + loop.getEnergy();
            childs.add(new Node(tree, this, level+1, loop, newStackedEnergy));
        }
        
        // recursion; do the same for each child that is just added
        if(level < tree.getTop().length() - 2 && 
           level < tree.getBottom().length() - 2) {
            for(Node child : childs) {
                child.addChilds();
            }
        }
    }
    
    /**
     * This method builds the template string that is used to retrieve the 
     * stacks that are possible childs of this node. The template is build based
     * on the top and bottom templates as given by tree and based on the stack 
     * that is represented by this node (because the child stack should 'fit' 
     * the stack of this node).
     * @return The template which should match the child nodes.
     */
    private String buildChildTemplate() {

        String result = "";
        
        // the last two nucleotides of this nodes' stack are the start of the
        // template for the next stack (to make them fit together)    
        if(loop != null) {
            String s = loop.name().toLowerCase();
            result += "" + s.charAt(3) + s.charAt(2);
        }
        // the root node doesn't represent a stack, so then we just take the 
        // correct part of the template as given in tree
        else {
            result += "" + tree.getTop().charAt(level) 
                           + tree.getBottom().charAt(level);
        }
        // the last two places are determined by the templates in tree
        result += "" + tree.getBottom().charAt(level+1) 
                       + tree.getTop().charAt(level+1);

        return result;
    }
    
    /**
     * This method recusively retrieves the top sequence of a hairpin 
     * (represented by a branch of the tree). The tree is traversed from bottom
     * to top over one branch getting the last of the two top nucleotides per 
     * stack.
     * @return The sequence at the top of the hairpin.
     */
    private String getTopSequence() {
        if(!parent.isRoot()) {
            return "" + parent.getTopSequence() + loop.name().toLowerCase().charAt(3);
        }
        else {
            return "" + loop.name().toLowerCase().charAt(0) + loop.name().toLowerCase().charAt(3);
        }
    }
    
    /**
     * This method recusively retrieves the bottom sequence of a hairpin 
     * (represented by a branch of the tree). The tree is traversed from bottom
     * to top over one branch getting the last of the two bottom nucleotides per 
     * stack.
     * @return The sequence at the bottom of the hairpin.
     */
    private String getBottomSequence() {
        if(!parent.isRoot()) {
            return "" + parent.getBottomSequence() + loop.name().toLowerCase().charAt(2);
        }
        else {
            return "" + loop.name().toLowerCase().charAt(1) + loop.name().toLowerCase().charAt(2);
        }
    }
    
    /**
     * This method returns the sequences of the hairpins that are the result 
     * after building this tree. This method should be called by the root node
     * to obtain all the found hairpins.
     * The results are obtained by recursively traversing the tree from top to
     * bottom. As soon as a leaf node is found that is at the correct level 
     * (length of the hairpin - 1) we know this is a possible design of the 
     * hairpin and the sequence of this hairpin (represented by the branch to
     * which this leaf belongs) is added to result.
     * @param result An array in which the resulting hairpin designs are stored.
     * @param loop The hairpin loop to add to the stacking region to make a 
     *             whole hairpin sequence out of it.
     */
    public void getResults(ArrayList<String> result, String loop) {
        for(Node child : childs) {
            if(child.isResultLeaf()) {
                StringBuffer start = new StringBuffer(child.getBottomSequence());
                start.reverse();
                String end = child.getTopSequence();
                result.add(start.toString() + loop + end);
            }
            else {
                // recursion
                child.getResults(result, loop);
            }
        }
    }
    
    /**
     * Returns if this node is the root node.
     * @return True when this node is the root of the tree.
     */
    private boolean isRoot() {
        return parent == null;
    }
    
    /**
     * Returns if this node is a leaf node (has no childs).
     * @return True if this node is a leaf node.
     */
    private boolean isLeaf() {
        return childs.isEmpty();
    }
    
    /**
     * Returns if the leaf node is part of a complete hairpin structure.
     * @return True if this node is a leaf node that is at the end of a 
     * complete hairpin.
     */
    private boolean isResultLeaf() {
    	int dest = Math.min(tree.getTop().length(), tree.getBottom().length());
        return isLeaf() && level == dest -1;
    }

    private int level;
    private Loop loop;
    private int stackedEnergy;
    private Tree tree;
    private Node parent;
    private ArrayList<Node> childs = new ArrayList<Node>();

}
