/**
 *  Copyright (C) Bastiaan van den Berg
 *
 *  This file is part of RNAHairpinDesigner.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.JScrollPane;
import javax.swing.border.TitledBorder;

/**
*
* 
*
* <p>
* This program can be used to design RNA sequences that fit a given template
* sequence and that form RNA hairpin structures with a defined stability 
* profile. 
* <p>
* An RNA hairpin structure consists of loops. A hairpin loop at the 
* left and a number of loops to the right that form the helix region. In this
* implementation the helix region only contains stack loops and no bulge or
* internal loops.
* <pre>
*      a
*   g     c - u - g - u - u - c  3'
*   |     |   |   |   |   |   |
*   u     g - a - c - a - g - g  5'
*      g
* </pre>
* 
* <p>
* See also http://2008.igem.org/Team:TUDelft/Temperature_software
* 
* @see Loop
* @see BasePair
* @see Nucleotide
* @see Tree
* @see Node
* @author Bastiaan van den Berg
*/
public class RNAHairpinDesigner extends JFrame implements ActionListener {

	/**
	 * Create new frame.
	 * @param name
	 */
	private RNAHairpinDesigner(String name) {
		super(name);
        //setResizable(false);
	}
	
	/**
	 * Add components to the pane.
	 * @param pane
	 */
	private void addComponentsToPane(final Container pane) {
		
		pane.setLayout(null);

		JPanel template = templatePanel();
		JPanel profile = profilePanel();
		JPanel result = resultPanel();
		
		int width = 400;
        template.setPreferredSize(new Dimension(width, 60));
        profile.setPreferredSize(new Dimension(width, 120));
        result.setPreferredSize(new Dimension(width, 190));

        Insets insets = pane.getInsets();
        Dimension size = template.getPreferredSize();
        template.setBounds(10 + insets.left, 10 + insets.top, size.width, size.height);
        size = profile.getPreferredSize();
        profile.setBounds(10 + insets.left, 75 + insets.top, size.width, size.height);
        size = result.getPreferredSize();
        result.setBounds(10 + insets.left, 200 + insets.top, size.width, size.height);

        pane.add(template);
        pane.add(profile);
        pane.add(result);
    }
	
	/**
	 * Returns the Template panel.
	 * @return Template panel.
	 */
	private JPanel templatePanel() {
		
        final JPanel templatePanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        templatePanel.setBorder(new TitledBorder("RNA Hairpin sequence template"));
        
        // 5 template components
        JLabel labelFiveEnd = new JLabel("5'");
        JLabel labelThreeEnd = new JLabel("3'");
        topTextField = new JTextField("gagga.uacuagaug", 16);
        topTextField.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
        bottomTextField = new JTextField("...............", 16);
        bottomTextField.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
        loopTextField = new JTextField("cuug", 5);
        loopTextField.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
        
        // add the components (flow left to right)
        templatePanel.add(labelFiveEnd);
        templatePanel.add(bottomTextField);
        templatePanel.add(loopTextField);
        templatePanel.add(topTextField);
        templatePanel.add(labelThreeEnd);
        
		return templatePanel;
	}

	/**
	 * Returns the profile panel.
	 * @return Profile panel.
	 */
	private JPanel profilePanel() {
		
		final JPanel profilePanel = new JPanel(null);
		profilePanel.setBorder(new TitledBorder("RNA hairpin stability profile"));
		
		// components
		profileTextArea = new JTextArea("-300, -300, -300, -300, -433, -566, -699, -832, -965, -1098, -1231, -1364, -1497, -1630, -1763, -1896", 3, 50);
		JLabel label = new JLabel("Maximal distance");
		distanceTextField = new JTextField("200", 5);
		
		// set fonts
		profileTextArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
		distanceTextField.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
		
		// make profile text area scrollable
		JScrollPane scrollPane = new JScrollPane(profileTextArea);
		
		// position components
		Insets insets = profilePanel.getInsets();
		scrollPane.setBounds(insets.left + 5, insets.top + 5, 380, 50);
		label.setBounds(insets.left + 5, insets.top + 65, 100, 20);
		distanceTextField.setBounds(insets.left + 110, insets.top + 65, 50, 20);
		
		// add the stuff to the profile panel
		profilePanel.add(scrollPane);
		profilePanel.add(label);
		profilePanel.add(distanceTextField);
		
		return profilePanel;
	}
	
	/**
	 * Returns the result panel.
	 * @return Result panel.
	 */
	private JPanel resultPanel() {
		
		final JPanel resultPanel = new JPanel(null);
		resultPanel.setBorder(new TitledBorder("Results"));
		
		// components
		JButton buttonResult = new JButton("Get Sequences");
		resultLabel = new JLabel("Number of results: ");
		resultTextArea = new JTextArea(6, 50);
		resultTextArea.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
		JButton buttonSave = new JButton("Save result");
		
		//
		buttonResult.setActionCommand("getSequences");
		buttonResult.addActionListener(this);
		buttonSave.setActionCommand("save");
		buttonSave.addActionListener(this);

		// make result text area scrollable
		scrollResultTextArea = new JScrollPane(resultTextArea, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		
		// position components
		Insets insets = resultPanel.getInsets();
		buttonResult.setBounds(insets.left + 5,	insets.top + 5, 120, 20);
		resultLabel.setBounds(insets.left + 130, insets.top + 5, 200, 20);
		scrollResultTextArea.setBounds(insets.left + 5, insets.top + 35, 380, 90);
		buttonSave.setBounds(insets.left + 5, insets.top + 135, 100, 20);
		
		// add the stuff to the result panel
		resultPanel.add(buttonResult);
		resultPanel.add(resultLabel);
		resultPanel.add(scrollResultTextArea);
		resultPanel.add(buttonSave);
		
		return resultPanel;
	}
	
	/**
	 * Designs the hairpin sequences when the getSequences button is pressed.
	 * Opens a save file dialog when the save button is pressed.
	 */
	public void actionPerformed(ActionEvent e) {
		if("getSequences".equals(e.getActionCommand())) {
			getHairpins();
		}
		if("save".equals(e.getActionCommand())) {
			JFileChooser fileChooser = new JFileChooser();
			//FileNameExtensionFilter filter = new FileNameExtensionFilter("txt");
			//fileChooser.setFileFilter(filter);
			if(fileChooser.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
				File file = fileChooser.getSelectedFile();
				if (file.getName().indexOf('.')==-1){
			        file = new File(file.getAbsolutePath()+".txt");
				}	 
				try {
					FileWriter fw = new FileWriter(file);
					fw.write(resultTextArea.getText());
					fw.flush();
					fw.close();
				}
				catch(IOException ioe) {
					// show error popup
					ioe.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * Checks the user input and uses it to design the hairpin sequences. The 
	 * resulting sequences are displayed in a textfield.
	 */
    private void getHairpins() {

    	// get template input
        String templateTop = topTextField.getText().trim();
        String templateBottom = bottomTextField.getText().trim();
        String loop = loopTextField.getText().trim();
        
        // check template input
        if(!correctStemTemplate(templateTop) || 
           !correctStemTemplate(templateBottom)) {
        	showMessage("The stem sequences may only contain the characters" +
        			" [aucg.].");
        	return;
        }
        if(!correctLoopTemplate(loop)) {
        	showMessage("The loop sequences may only contain the characters" +
        			" [aucg].");
        	return;
        }
        if(!(templateTop.length() == templateBottom.length())) {
        	showMessage("Both stem sequences should have the same length." +
        			System.getProperty("line.separator") +
        			System.getProperty("line.separator") +
        			"The current length of the first stem sequence is " +
        			bottomTextField.getText().length() + "." +
        			System.getProperty("line.separator") +
        			"The current length of the second stem sequence is " +
        			topTextField.getText().length() + ".");
        	return;
        }
        
        // get profile input
        int maxDist = 0;
        try {
        	maxDist = Integer.valueOf(distanceTextField.getText().trim());
        }
        catch(NumberFormatException e) {
        	showMessage("Maximal distance must be an integer.");
        	return;
        }
        if(maxDist <= 0) {
        	showMessage("Maximal distance must be greater than zero.");
        	return;
        }
        String[] pro = profileTextArea.getText().split(",");
        int[] energyProfile = new int[pro.length];
        int index = 0;
        for(String s : pro) {
        	s = s.trim();
        	try {
        		energyProfile[index] = Integer.valueOf(s);
        	}
        	catch(NumberFormatException e) {
        		showMessage("The stability profile must be a comma-seperated" +
        				"list containing only integers");
            	return;
        	}
        	index++;
        }
        if((energyProfile.length < topTextField.getText().length() - 1)) {
        	showMessage("the number of integers in the stability profile may" +
        			System.getProperty("line.separator") +
        			"not be more than one less than the length of a stem sequence." +
        			System.getProperty("line.separator") +
        			System.getProperty("line.separator") +
        			"current length of the stem sequence: " + 
        			topTextField.getText().length() +
        			System.getProperty("line.separator") +
        			"current number of integers within the energy profile: " +
        			energyProfile.length +
        			System.getProperty("line.separator") +
        			System.getProperty("line.separator") +
        			"The stability profile should contain at least " +
        			(topTextField.getText().length() - 1) + " values.");
        	return;
        }
        
        // build the tree
        Tree tree = new Tree(templateTop,templateBottom,energyProfile,maxDist);
        tree.buildTree();
        
        // get the resulting hairpin sequences from the tree
        ArrayList<String> result = tree.getResults(loop);
        
        // display the results in the result text area 
        String res = "";
        int i = 0;
        for(String s : result) {
            res += ">" + i + System.getProperty("line.separator");
            res += s + System.getProperty("line.separator");
            i++;
        }
        resultLabel.setText("Number of results: " + result.size());
        resultTextArea.setText(res);
    }
    
    /**
     * Checks if the provided template for the top or bottom sequence are
     * correct, containing only the characters .,a,c,g, and u.
     * @param template The template that has to be checked.
     * @return True if the template is correct.
     */
    private boolean correctStemTemplate(String template) {
    	Pattern pattern = Pattern.compile("[acgu. ]*", Pattern.CASE_INSENSITIVE);
    	Matcher matcher = pattern.matcher(template);
    	return matcher.matches();
    }
    
    /**
     * Checks if the provided template for the top or bottom sequence are
     * correct, containing only the characters a,c,g, and u.
     * @param template The template that has to be checked.
     * @return True if the template is correct.
     */
    private boolean correctLoopTemplate(String template) {
    	Pattern pattern = Pattern.compile("[acgu ]*", Pattern.CASE_INSENSITIVE);
    	Matcher matcher = pattern.matcher(template);
    	return matcher.matches();
    }
    
    /**
     * Shows a dialog with the given message.
     * @param message The message to be displayed in the dialog.
     */
    private void showMessage(String message) {
    	JOptionPane.showMessageDialog(this, message, "Message", 
    			JOptionPane.INFORMATION_MESSAGE);
    }
    
    /**
     * Create the GUI and show it.  For thread safety,
     * this method is invoked from the
     * event dispatch thread.
     */
    private static void createAndShowGUI() {
        //Create and set up the window.
        RNAHairpinDesigner frame = new RNAHairpinDesigner("RNA hairpin designer");
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        //Set up the content pane.
        frame.addComponentsToPane(frame.getContentPane());
        
        // Size and display the window.
        Insets insets = frame.getInsets();
        frame.setSize(430 + insets.left + insets.right,
                      430 + insets.top + insets.bottom);

        //Display the window.
        //frame.pack();
        frame.setVisible(true);
    }
    
    /**
     * Set up the GUI and run the program.
     * @param args
     */
    public static void main(String[] args) {
        /* Use an appropriate Look and Feel */
        try {
            //UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
            UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");
        } catch (UnsupportedLookAndFeelException ex) {
            ex.printStackTrace();
        } catch (IllegalAccessException ex) {
            ex.printStackTrace();
        } catch (InstantiationException ex) {
            ex.printStackTrace();
        } catch (ClassNotFoundException ex) {
            ex.printStackTrace();
        }
        /* Turn off metal's use of bold fonts */
        UIManager.put("swing.boldMetal", Boolean.FALSE);
        
        //Schedule a job for the event dispatch thread:
        //creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                createAndShowGUI();
            }
        });
    }
    
    // template
    private JTextField bottomTextField;
    private JTextField topTextField;
    private JTextField loopTextField;
    // profile
    private JTextArea profileTextArea;
    private JTextField distanceTextField;
    // results
    private JTextArea resultTextArea;
    private JScrollPane scrollResultTextArea;
    private JLabel resultLabel;
    
    /**
	 * 
	 */
	private static final long serialVersionUID = 1742898806019246237L;
}
