/**
 *  Copyright (C) Bastiaan van den Berg
 *
 *  This file is part of RNAHairpinDesigner.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Library General Public License as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;

/**
 * This class represents a tree which is used as a data structure that stores 
 * multiple RNA hairpins. A branch within the tree represents a hairpin 
 * structure that is build up by a number of loops, which are the nodes. 
 * Each hairpin within the tree must have a stability profile that stays within
 * the boundaries of the energyProfile and have a sequence that matches the 
 * given template sequence.
 * 
 * <p>
 * See also http://2008.igem.org/Team:TUDelft/Temperature_software
 * 
 * @see Node
 * @author Bastiaan van den Berg 
 */
public class Tree {

	/**
	 * Creates a new tree.
	 * @param top A template for the top sequence of the hairpin.
	 * @param bottom A template for the bottom sequence of the hairpin.
	 * @param energyProfile The energy profile.
	 * @param maxDist The maximal allowed distance to the energy profile.
	 */
    public Tree(String top, String bottom, int[] energyProfile, int maxDist) {
        this.top = top;
        this.bottom = bottom;
        this.energyProfile = energyProfile;
        this.maxDist = maxDist;
        root = new Node(this, null, 0, null, 0);
    }
    
    /**
     * Recursively builds up the tree.
     * @see Node#addChilds()
     */
    public void buildTree() {
        root.addChilds();
    }
    
    /**
     * Returns the resulting RNA hairpins that fit the template and have a 
     * stability profile that fits the given energy profile.
     * @param loop The sequense of the hairpin loop that connects the top and
     *             bottom sequences.
     * @return List with RNA sequences.
     */
    public ArrayList<String> getResults(String loop) {
        ArrayList<String> result = new ArrayList<String>();
        root.getResults(result, loop);
        return result;
    }

    /**
     * Returns the top RNA sequence.
     * @return The top RNA sequence.
     */
    public String getTop() {
        return top;
    }
    
    /**
     * Returns the bottom RNA sequence.
     * @return The bottom RNA sequence.
     */
    public String getBottom() {
        return bottom;
    }
    
    /**
     * Returns the energy profile.
     * @return The energy profile.
     */
    public int[] getEnergyProfile() {
        return energyProfile;
    }
    
    /**
     * Returns the maximal distance to the energy profile.
     * @return The maximal distance to the energy profile.
     */
    public int getMaxDist() {
        return maxDist;
    }

    private String top;
    private String bottom;
    private int[] energyProfile;
    private int maxDist;

    // The root node of the tree.
    private Node root;
}
